import re
import datetime

# A dictionary of filter classes, keyed by their name attributes.
FILTER_REGISTRY = {}


def _now():
    """
    Mockable proxy to ``datetime.datetime.now()``
    """
    return datetime.datetime.now()


class FilterRegistry(type):
    """
    Metaclass used to automatically register new filter classes in our filter
    registry. Enables shorthand filter notation.

        >>> from es_fluent.builder import QueryBuilder
        >>> query_builder = QueryBuilder()
        >>> query_builder.add_filter('missing', 'boop').to_query()['filter']
        {'missing': {'field': 'field_name'}}
    """
    def __new__(cls, clsname, bases, attrs):
        newclass = super(FilterRegistry, cls).__new__(
            cls, clsname, bases, attrs)
        register_filter(newclass)
        return newclass


def register_filter(filter_cls):
    """
    Adds the ``filter_cls`` to our registry.
    """
    if filter_cls.name is None:
        return
    elif filter_cls.name in FILTER_REGISTRY:
        raise RuntimeError(
            "Filter class already registered: {}".format(filter_cls.name))
    else:
        FILTER_REGISTRY[filter_cls.name] = filter_cls


def build_filter(filter_or_string, *args, **kwargs):
    """
    Overloaded filter construction. If ``filter_or_string`` is a string
    we look up it's corresponding class in the filter registry and return it.
    Otherwise, assume ``filter_or_string`` is an instance of a filter.

    :return: :class:`~es_fluent.filters.Filter`
    """
    if isinstance(filter_or_string, basestring):
        # Names that start with `~` indicate a negated filter.
        if filter_or_string.startswith('~'):
            filter_name = filter_or_string[1:]
            return ~FILTER_REGISTRY[filter_name](*args, **kwargs)
        else:
            filter_name = filter_or_string
            return FILTER_REGISTRY[filter_name](*args, **kwargs)
    else:
        return filter_or_string


class Filter(object):
    """
    The base filter.
    """

    #: The shorthand name of the filter.
    name = None

    #: Auto-register any Filter subclass with our registry.
    __metaclass__ = FilterRegistry

    def __invert__(self):
        """
        Returns this filter wrapped in a :class:`es_fluent.filters.Not` filter.
        :
        """
        not_filter = Not()
        not_filter.add_filter(self)
        return not_filter

    def to_query(self):
        """
        Serializes this ``Filter`` and any descendants into a json-serializable
        dictionary suitable for use with the elasticsearch api.
        """
        raise NotImplementedError()


class Generic(Filter):
    """
    Contains a generic list of filters. Serialized as a dictionary.
    """
    def __init__(self):
        self.filters = []

    def is_empty(self):
        """
        :return: ``True`` if this filter has nested clauses ``False``.
        """
        return len(self.filters) == 0

    def add_filter(self, filter_or_string, *args, **kwargs):
        """
        Appends a filter.
        """
        self.filters.append(build_filter(filter_or_string, *args, **kwargs))

        return self

    def and_filter(self, filter_or_string, *args, **kwargs):
        """
        Adds a list of :class:`~es_fluent.filters.And` clauses, automatically
        generating :class:`~es_fluent.filters.And` filter if it does not
        exist.
        """
        and_filter = self.find_filter(And)

        if and_filter is None:
            and_filter = And()
            self.filters.append(and_filter)

        and_filter.add_filter(build_filter(
            filter_or_string, *args, **kwargs))

        return and_filter

    def or_filter(self, filter_or_string, *args, **kwargs):
        """
        Adds a list of :class:`~es_fluent.filters.Or` clauses, automatically
        generating the an :class:`~es_fluent.filters.Or` filter if it does not
        exist.
        """
        or_filter = self.find_filter(Or)

        if or_filter is None:
            or_filter = Or()
            self.filters.append(or_filter)

        or_filter.add_filter(build_filter(
            filter_or_string, *args, **kwargs
        ))

        return or_filter

    def find_filter(self, filter_cls):
        """
        Find or create a filter instance of the provided ``filter_cls``. If it
        is found, use remaining arguments to augment the filter otherwise
        create a new instance of the desired type and add it to the
        current :class:`~es_fluent.builder.QueryBuilder` accordingly.
        """
        for filter_instance in self.filters:
            if isinstance(filter_instance, filter_cls):
                return filter_instance

        return None


class Dict(Generic):
    """
    Contains a generic dictionary of filters e.g. in a top level ES Query
    we may have::

        { "filtered": {"filter": {"and": {...}, "or": {...}, "exists": {...} }

    The Dict filter may represent the dictionary inside of "filtered.filter".
    """
    def to_query(self):
        """
        Iterates over all filters and converts them to an Elastic HTTP API
        suitable query.

        Note: each :class:`~es_fluent.filters.Filter` is free to set it's own
        filter dictionary. ESFluent does not attempt to guard against filters
        that may clobber one another.  If you wish to ensure that filters are
        isolated, nest them inside of a boolean filter such as
        :class:`~es_fluent.filters.And` or :class:`~es_fluent.filters.Or`.
       """
        query = {}
        for filter_instance in self.filters:
            filter_query = filter_instance.to_query()
            query.update(filter_query)

        return query


class And(Generic):
    """
    A multi-clause filter that ands's all sub-filters added to it.

    `Elastic docs <https://www.elastic.co/guide/en/elasticsearch/reference/1.7/query-dsl-and-filter.html>`_.
    """
    name = 'and'

    def to_query(self):
        clauses = []
        for filter_instance in self.filters:
            clauses.append(filter_instance.to_query())
        return {
            "and": clauses
        }


class Or(Generic):
    """
    A multi-clause filter that ``ors`` all sub-filters added to it.

    `Elastic docs <https://www.elastic.co/guide/en/elasticsearch/reference/1.7/query-dsl-or-filter.html>`_.
    """
    name = 'or'

    def to_query(self):
        clauses = []
        for filter_instance in self.filters:
            clauses.append(filter_instance.to_query())
        return {
            "or": clauses
        }


class Not(Dict):
    """
    A filter that inverts it's clauses.

    `Elastic docs <https://www.elastic.co/guide/en/elasticsearch/reference/1.7/query-dsl-not-filter.html>`_.
    """
    name = 'not'

    def to_query(self):
        query = super(Not, self).to_query()
        return {
            "not": query
        }


class Terminal(Filter):
    """
    A filter that cannot contain nested filters. Merge behavior is to clobber
    existing clauses rather than appending additional clauses.
    """
    def add_filter(self, filter_instance):
        """
        As a :class:`~es_fluent.filters.Terminal` filter, adding nested filters
        is not allowed.
        """
        raise RuntimeError("Terminal filters do not support sub-filters")

    def find_filter(self, filter_cls):
        """
        As a :class:`~es_fluent.filters.Terminal` filter, adding nested filters
        is not allowed and therefore, finding sub-filters is not supported.
        """
        raise RuntimeError("Terminal filters do not support sub-filters")


class Custom(Terminal):
    """
    Allows for an entirely custom dictionary to be used and passed verbatim
    when `to_query` is invoked.
    """
    name = 'custom'

    def __init__(self, query):
        self.query = query

    def to_query(self):
        return self.query


class Exists(Terminal):
    """
    Checks whether a field exists in the source document.

    `Elastic docs <https://www.elastic.co/guide/en/elasticsearch/reference/1.7/query-dsl-exists-filter.html>`_.
    """
    name = 'exists'

    def __init__(self, value):
        self.value = value

    def to_query(self):
        return {
            "exists": {
                "field": self.value
            }
        }


class Range(Dict):
    """
    A Filter for ranges of values, supporting ``lt``, ``lte``, ``gt``, ``gte``
    comparisons.

    `Elastic docs <https://www.elastic.co/guide/en/elasticsearch/reference/1.7/query-dsl-range-filter.html>`_.
    """
    name = 'range'

    def __init__(self, name, gte=None, lte=None, lt=None, gt=None):
        self.name = name
        self.gte = gte
        self.lte = lte
        self.lt = lt
        self.gt = gt

    def to_query(self):
        criteria = {}
        for comparator in ['lte', 'lt', 'gt', 'gte']:
            if getattr(self, comparator) is not None:
                criteria[comparator] = getattr(self, comparator)

        return {
            "range": {
                self.name: criteria
            }
        }


class Age(Dict):
    """
    Similar to a range filter. Operates on times. When querified, we convert
    the age in seconds into a datetime relative to the current
    ``datetime.datetime.now()``.
    """
    name = 'age'

    def __init__(self, name, gte=None, lte=None, lt=None, gt=None):
        self.name = name
        # Yes, this flipping of gt/lt is deliberate. Saying something is
        # less than 3600 seconds old is equivalent to saying it's timestamp
        # is greater that `now` - 3600.
        self.lte = float(gte) if gte is not None else None
        self.gte = float(lte) if lte is not None else None
        self.gt = float(lt) if lt is not None else None
        self.lt = float(gt) if gt is not None else None

    def to_query(self):
        criteria = {}
        now = _now()
        for comparator in ['lte', 'lt', 'gt', 'gte']:
            if getattr(self, comparator) is not None:
                seconds = getattr(self, comparator)
                age = datetime.timedelta(seconds=seconds)
                value = (now - age).isoformat()
                criteria[comparator] = value
        return {
            "range": {
                self.name: criteria
            }
        }


class Term(Terminal):
    """
    `Elastic docs <https://www.elastic.co/guide/en/elasticsearch/reference/1.7/query-dsl-term-filter.html>`_.
    """
    name = 'term'

    def __init__(self, name, value):
        self.name = name
        self.value = value

    def to_query(self):
        return {
            "term": {
                self.name: self.value
            }
        }


class Missing(Terminal):
    """
    Filters documents that to do not contain a given field.

    `Elastic docs <https://www.elastic.co/guide/en/elasticsearch/reference/1.7/query-dsl-term-filter.html>`_.
    """
    name = 'missing'

    def __init__(self, name):
        self.name = name

    def to_query(self):
        return {
            "missing": {
                "field": self.name
            }
        }


class Terms(Terminal):
    """
    Matches documents that contain multiple exact values.

    `Elastic docs <https://www.elastic.co/guide/en/elasticsearch/reference/1.7/query-dsl-terms-filter.html>`_.
    """
    name = 'terms'

    def __init__(self, name, values):
        self.name = name
        self.values = values

    def to_query(self):
        return {
            "terms": {
                self.name: self.values
            }
        }


class RegExp(Terminal):
    """
    Matches documents whose given field matches a provided regular expression.

    `Elastic docs <https://www.elastic.co/guide/en/elasticsearch/reference/1.7/query-dsl-terms-filter.html>`_.
    """
    name = 'regexp'

    def __init__(self, name, expression):
        # Immediately try to compile the expression to check it's validity.
        re.compile(expression)
        self.name = name
        self.expression = expression

    def to_query(self):
        return {
            "regexp": {
                self.name: self.expression
            }
        }


class Script(Generic):
    """
    Represents script criteria.

    `Elastic docs <https://www.elastic.co/guide/en/elasticsearch/reference/1.7/query-dsl-script-filter.html>`_.
    """
    name = 'script'

    def to_query(self):
        query = {}

        for filter_instance in self.filters:
            filter_query = filter_instance.to_query()
            query.update(filter_query)

        return query


class ScriptID(Terminal):
    """
    Represents a pre-indexed script filter.

    `Elastic docs <https://www.elastic.co/guide/en/elasticsearch/reference/1.7/query-dsl-script-filter.html>`_.
    """
    name = 'script_id'

    def __init__(self, name, script_id, script_params, lang='groovy'):
        self.name = name
        self.script_id = script_id
        self.script_params = script_params
        self.lang = lang

    def to_query(self):
        return {
            self.name: {
                'lang': self.lang,
                'script_id': self.script_id,
                'params': self.script_params
            }
        }



class ScriptFields(object):
    """
    Represents a collection of requested script fields.
    """
    def __init__(self):
        self.fields = []

    def add_field(self, field_instance):
        """
        Appends a field.
        """
        if isinstance(field_instance, BaseScriptField):
            field_instance = field_instance
        else:
            raise ValueError('Expected a basetring or Field instance')

        self.fields.append(field_instance)

        return self

    def is_empty(self):
        """
        Returns True if there are no script fields, False otherwise.
        """
        return len(self.fields) == 0

    def to_query(self):
        """
        Returns a json-serializable representation.
        """
        query = {}

        for field_instance in self.fields:
            query.update(field_instance.to_query())

        return query


class BaseScriptField(object):
    """
    Manages a basic field. The intention here is to give script_fields
    and non script fields a similar taxonomy.
    """


class ScriptField(BaseScriptField):
    """
    Represents a script field.
    """
    def __init__(self, name, script, lang='groovy', **kwargs):
        """
        :param name: The resulting name of the field.
        :param script: The script for the script field.
        :param lang: The language of the script.
        :param kwargs: Additional keyword arguments become the params for
            the script.
        """
        self.name = name
        self.script = script
        self.lang = lang
        self.script_params = kwargs

    def to_query(self):
        """
        Returns a json-serializable representation.
        """
        return {
            self.name: {
                'lang': self.lang,
                'script': self.script,
                'params': self.script_params
            }
        }


class ScriptIDField(BaseScriptField):
    """
    Represents a pre-indexed script field.
    """
    def __init__(self, name, script_id, lang='groovy', **kwargs):
        """
        :param name: The resulting name of the field.
        :param script_id: The id of the pre-indexed script field.
        :param lang: The language of the script.
        :param kwargs: Additional keyword arguments become args for the script.
        """
        self.name = name
        self.script_id = script_id
        self.lang = lang
        self.script_params = kwargs

    def to_query(self):
        """
        Returns a json-serializable representation.
        """
        return {
            self.name: {
                'lang': self.lang,
                'script_id': self.script_id,
                'params': self.script_params
            }
        }



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_filters
----------------------------------

Tests for `filters` module.
"""

import unittest

from es_fluent.filters.geometry import GeoJSON, IndexedShape


class TestEs_fluent_filters(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_geoshape_point(self):
        geojson = {
            "type": "Point",
            "coordinates": [0, 0]
        }
        geoshape = GeoJSON("geom", geojson)
        self.assertEquals(geoshape.to_query(), {
            "geo_shape": {
                "geom": {
                    "shape": {
                        "type": "Point",
                        "coordinates": [0, 0]
                    }
                }
            }
        })

    def test_geoshape_feature(self):
        geojson = {
            "type": "Feature",
            "properties": {
                "hello": "I should be removed"
            },
            "geometry": {
                "type": "Point",
                "coordinates": [0, 0]
            }
        }
        geoshape = GeoJSON("geom", geojson)
        self.assertEquals(geoshape.to_query(), {
            "geo_shape": {
                "geom": {
                    "shape": {
                        "type": "Point",
                        "coordinates": [0, 0]
                    }
                }
            }
        })

    def test_geoshape_feature_collection(self):
        geojson = {
            "type": "FeatureCollection",
            "features": [{
                "type": "Feature",
                "properties": {
                    "hello": "I should be removed"
                },
                "geometry": {
                    "type": "Point",
                    "coordinates": [0, 0]
                }
            }]
        }
        geoshape = GeoJSON("geom", geojson)
        self.assertEquals(geoshape.to_query(), {
            "geo_shape": {
                "geom": {
                    "shape": {
                        "type": "GeometryCollection",
                        "geometries": [{
                            "type": "Point",
                            "coordinates": [0, 0]
                        }]
                    }
                }
            }
        })

    def test_indexed_shape(self):
        indexedshape = IndexedShape('geom', 'shape_id', 'index_name',
                                    'doc_type', 'path')
        self.assertEquals(indexedshape.to_query(), {
            "geo_shape": {
                "geom": {
                    "indexed_shape": {
                        "id": "shape_id",
                        "index": "index_name",
                        "type": "doc_type",
                        "path": "path"
                    }
                }
            }
        })


if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())



"""
Geometry related filters require additional dependencies.  Hence they're broken
out into their own module.
"""
from copy import deepcopy
from . import Terminal


def prepare_geojson(geojson):
    """
    Modifies incoming GeoJSON to make it Elastic friendly. This means:

        1. CW orientation of polygons.
        2. Re-casting of Features and FeatureCollections to Geometry and
           GeometryCollections.
    """
    # TODO CW orientation.
    geojson = deepcopy(geojson)

    if geojson["type"] == "Feature":
        geojson = geojson["geometry"]
        if hasattr(geojson, 'properties'):
            del geojson['properties']

    if geojson["type"] == "FeatureCollection":
        geojson["type"] = "GeometryCollection"
        geojson["geometries"] = [
            feature["geometry"] for feature in geojson["features"]
        ]
        del geojson["features"]

    return geojson


class GeoJSON(Terminal):
    """
    Manages querying by GeoJSON. Automatically converts incoming GeoJSON
    to elasticsearch friendly geometry. This generally means::

    #. CW orientation of polygons.
    #. Re-casting of Features and FeatureCollections to Geometry and
       GeometryCollections.
    """
    name = 'geometry'

    def __init__(self, name, geojson):
        self.name = name
        self.geojson = prepare_geojson(geojson)

    def to_query(self):
        """
        Returns a json-serializable representation.
        """
        return {
            "geo_shape": {
                self.name: {
                    "shape": self.geojson
                }
            }
        }


class IndexedShape(Terminal):
    """
    Searches by a previously indexed Geometry.
    """
    name = "indexed_geometry"

    def __init__(self, name, shape_id, index_name, doc_type, path):
        """
        :param string name: The field to match against the target shape.
        :param string shape_id: The id of the indexed shape within the index.
        :param string index_name: The name of the index containing our shape.
        :param string doc_type: The type of document within index_name.
        :param string path:
            The location of geometry field within the indexed doc.
        """
        self.name = name
        self.shape_id = shape_id
        self.index_name = index_name
        self.doc_type = doc_type
        self.path = path

    def to_query(self):
        """
        Returns a json-serializable representation.
        """
        return {
            "geo_shape": {
                self.name: {
                    "indexed_shape":  {
                        "index": self.index_name,
                        "type": self.doc_type,
                        "id": self.shape_id,
                        "path": self.path
                    }
                }
            }
        }



"""
Utilized by the :class:`~es_fluent.builder.QueryBuilder` to optionally
restrict the fields returned by an Elastic query.
"""


class Fields(object):
    """
    Represents a collection of fields to be requested at query time.
    """
    def __init__(self):
        self.fields = []

    def add_field(self, field_instance_or_string):
        """
        Appends a field, can be a :class:`~es_fluent.fields.Field` or string.
        """
        if isinstance(field_instance_or_string, basestring):
            field_instance = Field(field_instance_or_string)
        elif isinstance(field_instance_or_string, Field):
            field_instance_or_string = field_instance
        else:
            raise ValueError('Expected a basetring or Field instance')

        self.fields.append(field_instance)

        return self

    def to_query(self):
        """
        Serializes this into a json-serializable dictionary suitable for use
        with the elasticsearch api.
        """
        return [field.name for field in self.fields]

    def is_empty(self):
        """
        Returns ``False`` if no fields have been added, ``True`` otherwise.
        """
        return len(self.fields) == 0


class Field(object):
    """
    Manages a basic field. These occur inside of an elasticsearch fields array.
    """
    def __init__(self, name):
        self.name = name



from es_fluent.filters import Dict
from es_fluent.script_fields import ScriptFields
from es_fluent.fields import Fields


class QueryBuilder(object):
    def __init__(self):
        self.root_filter = Dict()
        self.script_fields = ScriptFields()
        self.fields = Fields()

        self.sorts = []
        self.source = True
        self._size = None

    @property
    def size(self):
        """
        Sets current size limit of the ES response, which limits the number of
        documents returned. By default this is unset and the number of
        documents returned is up to ES.

        :return:
            The current size limit.
        """
        return self._size

    @size.setter
    def size(self, size):
        """
        Sets the size of the ES response.

        :param size: The number of documents to limit the response to.
        """
        self._size = size

    def and_filter(self, filter_or_string, *args, **kwargs):
        """
        Convenience method to delegate to the root_filter to generate an
        :class:`~es_fluent.filters.And` clause.

        :return: :class:`~es_fluent.builder.QueryBuilder`
        """
        self.root_filter.and_filter(filter_or_string, *args, **kwargs)
        return self

    def or_filter(self, filter_or_string, *args, **kwargs):
        """
        Convenience method to delegate to the root_filter to generate an `or`
        clause.

        :return: :class:`~es_fluent.builder.QueryBuilder`
        """
        self.root_filter.or_filter(filter_or_string, *args, **kwargs)
        return self

    def add_filter(self, filter_or_string, *args, **kwargs):
        """
        Adds a filter to the query builder's filters.

        :return: :class:`~es_fluent.builder.QueryBuilder`
        """
        self.root_filter.add_filter(filter_or_string, *args, **kwargs)
        return self

    def add_field(self, field_instance):
        """
        Adds a field to the query builder. The default behavior is
        to return all fields. Explicitly adding a single field will
        result in only that source field being returned.

        :return: :class:`~es_fluent.builder.QueryBuilder`
        """
        self.fields.add_field(field_instance)
        return self

    def add_script_field(self, field_instance):
        """
        Add a script field to the query. The `field_instance` should be
        an instance of `es_fluent.script_fields`.

        :return: :class:`~es_fluent.builder.QueryBuilder`
        """
        self.script_fields.add_field(field_instance)
        return self

    def find_filter(self, filter_cls):
        """
        Finds an existing filter using a filter class `filter_cls`. If not
        found, None is returned.

        This method is useful in cases where one wants to modify and extend
        and existing clause, a common example might be an
        :class:`~es_fluent.filters.And` filter. The method only looks in the
        query's top-level filter and does not recurse.

        :param: ``filter_cls``
            The the :class:`~es_fluent.filters.Filter` class
            to find.
        """
        return self.root_filter.find_filter(filter_cls)

    def to_query(self):
        result = {}

        if not self.root_filter.is_empty():
            result['filter'] = self.root_filter.to_query()

        if not self.script_fields.is_empty():
            result['script_fields'] = self.script_fields.to_query()

        if not self.fields.to_query():
            result['fields'] = self.fields.to_query()

        # We don't bother with reprensenting sort as an object.
        if len(self.sorts):
            result['sort'] = self.sorts

        if self._size is not None:
            result['size'] = self._size

        result['_source'] = self.source

        return result

    def disable_source(self):
        """
        Don't include ``_source`` document in results.

        :return: :class:`~es_fluent.builder.QueryBuilder`
        """
        self.source = False

        return self

    def enable_source(self):
        """
        Include ``_source`` document in results.

        :return: :class:`~es_fluent.builder.QueryBuilder`
        """
        self.source = True

    def sort(self, field, direction="asc"):
        """
        Adds sort criteria.
        """
        if not isinstance(field, basestring):
            raise ValueError("Field should be a string")
        if direction not in ["asc", "desc"]:
            raise ValueError("Sort direction should be `asc` or `desc`")

        self.sorts.append({field: direction})

    def remove_sort(self, field_name):
        """
        Clears sorting criteria affecting ``field_name``.
        """
        self.sorts = [dict(field=value) for field, value in self.sorts if field
                      is not field_name]

    def sort_reset(self):
        """
        Resets sorting criteria.
        """
        self.sorts = []



#!/usr/bin/env python
# -*- coding: utf-8 -*-


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read().replace('.. :changelog:', '')

requirements = []

test_requirements = [
    'sphinx_rtd_theme',
    'mock'
]

setup(
    name='es-fluent',
    version='0.0.2',
    description="Fluent API for generating elastic queries.",
    long_description=readme + '\n\n' + history,
    author="Jacob Straszynski",
    author_email='jacob.straszynski@planet.com',
    url='https://github.com/planetlabs/es_fluent',
    packages=[
        'es_fluent',
    ],
    package_dir={'es_fluent':
                 'es_fluent'},
    include_package_data=True,
    install_requires=requirements,
    license="Apache 2",
    zip_safe=False,
    keywords='es_fluent',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
    ],
    test_suite='tests',
    tests_require=test_requirements
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_es_fluent
----------------------------------

Tests for `es_fluent` module.
"""

import unittest
import datetime
from mock import patch

from es_fluent.builder import QueryBuilder
from es_fluent.filters import Exists, Or, And, Range, Term, Terms, ScriptID, \
    Age, Missing
from es_fluent.script_fields import ScriptField, ScriptIDField


class TestEs_fluent(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_fluent_constructor(self):
        query_builder = QueryBuilder()
        self.assertIsNotNone(query_builder)

    def test_empty_and_filter(self):
        and_filter = And()
        self.assertEquals({"and": []}, and_filter.to_query())

    def test_single_clause_and_filter(self):
        and_filter = And()
        and_filter.add_filter(Exists("value"))
        self.assertEquals({"and": [
            {"exists": {"field": "value"}}
        ]}, and_filter.to_query())

    def test_empty_or_filter(self):
        or_filter = Or()
        self.assertEquals({"or": []}, or_filter.to_query())

    def test_single_clause_or_filter(self):
        or_filter = Or()
        or_filter.add_filter(Exists("value"))
        self.assertEquals({"or": [
            {"exists": {"field": "value"}}
        ]}, or_filter.to_query())

    def test_query_builder_and_filter(self):
        query_builder = QueryBuilder()
        query_builder.and_filter(
            Exists('value')
        )
        self.assertEquals(
            query_builder.to_query()['filter'],
            {
                "and": [
                    {"exists": {"field": "value"}}
                ]
            }
        )

    def test_multi_and_filter(self):
        """
        When a given filter type already exists, we smartly append to it if
        possible rather than clobbering an existing filter.
        """
        query_builder = QueryBuilder()
        query_builder.and_filter(
            Exists('field_one')
        )
        query_builder.and_filter(
            Exists('field_two')
        )
        self.assertEquals(
            query_builder.to_query()['filter'],
            {
                "and": [
                    {"exists": {"field": "field_one"}},
                    {"exists": {"field": "field_two"}}
                ]
            }
        )

    def test_filter_clobber(self):
        query_builder = QueryBuilder()
        query_builder.add_filter(Exists('value'))
        self.assertEquals(
            query_builder.to_query()['filter'],
            {"exists": {"field": "value"}},
        )
        query_builder.add_filter(Exists('value_two'))
        self.assertEquals(
            query_builder.to_query()['filter'],
            {"exists": {"field": "value_two"}},
        )

    def test_nested_or_filter(self):
        """
        It should be possible to add another multi-clause filter inside
        of an existing multi-clause filter and have them nest correctly.
        """
        query_builder = QueryBuilder()
        query_builder.or_filter(
            Exists('value')
        )
        or_filter = query_builder.find_filter(Or)
        or_filter.or_filter(
            Exists('nested_value')
        )
        self.assertEquals(
            query_builder.to_query()['filter'],
            {
                "or": [
                    {"exists": {"field": "value"}},
                    {
                        "or": [
                            {"exists": {
                                "field": "nested_value"
                            }}
                        ]
                    }
                ]
            }
        )

    def test_nested_and_filter(self):
        """
        It should be possible to add another multi-clause filter inside
        of an existing multi-clause filter and have them nest correctly.
        """
        query_builder = QueryBuilder()
        query_builder.and_filter(
            Exists('value')
        )
        and_filter = query_builder.find_filter(And)
        and_filter.and_filter(
            Exists('nested_value')
        )
        self.assertEquals(
            query_builder.to_query()['filter'],
            {
                "and": [
                    {"exists": {"field": "value"}},
                    {
                        "and": [
                            {"exists": {
                                "field": "nested_value"
                            }}
                        ]
                    }
                ]
            }
        )

    def test_inversion_and(self):
        """
        Filters can return an inverted version of themselve using the `~`
        operator.
        """
        query_builder = QueryBuilder()
        query_builder.and_filter(
            ~Exists('and_not_value')
        )
        query_builder.and_filter(
            Exists('exists_value')
        )
        self.assertEquals(
            query_builder.to_query()['filter'],
            {
                "and": [
                    {"not": {"exists": {"field": "and_not_value"}}},
                    {"exists": {"field": "exists_value"}}
                ]
            }
        )

    def test_inversion_or(self):
        """
        Filters can return an inverted version of themselve using the `~`
        operator.
        """
        query_builder = QueryBuilder()
        query_builder.or_filter(
            ~Exists('and_not_value')
        )
        query_builder.or_filter(
            Exists('exists_value')
        )
        self.assertEquals(
            query_builder.to_query()['filter'],
            {
                "or": [
                    {"not": {"exists": {"field": "and_not_value"}}},
                    {"exists": {"field": "exists_value"}}
                ]
            }
        )

    def test_range_full(self):
        """
        All valid range comparators should be represented in the final query.
        """
        query_builder = QueryBuilder()
        query_builder.add_filter(Range(
            'field', lte=0.5, lt=0.6, gt=0.7, gte=0.8
        ))
        self.assertEquals(
            query_builder.to_query()['filter'],
            {
                "range": {
                    "field": {
                        "lte": 0.5,
                        "lt": 0.6,
                        "gt": 0.7,
                        "gte": 0.8
                    }
                }
            }
        )

    def test_range_partial(self):
        """
        If only a specific range criteria is provided, only it should appear
        in the final query.
        """
        query_builder = QueryBuilder()
        query_builder.add_filter(Range(
            'field', lte=0.5
        ))
        self.assertEquals(
            query_builder.to_query()['filter'],
            {
                "range": {
                    "field": {
                        "lte": 0.5,
                    }
                }
            }
        )

    def test_term(self):
        query_builder = QueryBuilder()
        query_builder.add_filter(Term(
            'field', 0.5
        ))
        self.assertEquals(
            query_builder.to_query()['filter'],
            {
                "term": {"field": 0.5}
            }
        )

    def test_terms(self):
        query_builder = QueryBuilder()
        query_builder.add_filter(Terms(
            'field', 0.5
        ))
        self.assertEquals(
            query_builder.to_query()['filter'],
            {
                "terms": {"field": 0.5}
            }
        )

    def test_script_id(self):
        query_builder = QueryBuilder()
        query_builder.add_filter(ScriptID(
            'name', 'script_id', {"key": 0.5}
        ))
        self.assertEquals(
            query_builder.to_query()['filter'],
            {
                'name': {
                    "script_id": "script_id",
                    "lang": "groovy",
                    "params": {
                        "key": 0.5
                    }
                }
            }
        )

    def test_multi_script_id(self):
        query_builder = QueryBuilder()
        query_builder.add_filter(ScriptID(
            'name', 'script_id', {"key": 0.5}
        ))
        query_builder.add_filter(ScriptID(
            'name_two', 'script_id_two', {"key_two": 0.4}
        ))
        self.assertEquals(
            query_builder.to_query()['filter'],
            {
                'name': {
                    "script_id": "script_id",
                    "lang": "groovy",
                    "params": {
                        "key": 0.5
                    }
                },
                'name_two': {
                    "script_id": "script_id_two",
                    "lang": "groovy",
                    "params": {
                        "key_two": 0.4
                    }
                }
            }
        )

    def test_script_field(self):
        query_builder = QueryBuilder()
        query_builder.add_script_field(ScriptField(
            'my_script',
            '_source.field * factor',
            factor=2
        ))
        self.assertEquals(query_builder.to_query()['script_fields'], {
            'my_script': {
                'lang': 'groovy',
                'script': '_source.field * factor',
                'params': {
                    'factor': 2
                }
            }
        })

    def test_script_id_field(self):
        query_builder = QueryBuilder()
        query_builder.add_script_field(ScriptIDField(
            'my_script',
            'my_script_id',
            factor=2
        ))
        self.assertEquals(query_builder.to_query()['script_fields'], {
            'my_script': {
                'lang': 'groovy',
                'script_id': 'my_script_id',
                'params': {
                    'factor': 2
                }
            }
        })

    def test_add_and_shorthand(self):
        query_builder = QueryBuilder()
        query_builder.and_filter('exists', 'some_field')
        query_builder.and_filter('range', 'range_field', lte=5)
        self.assertEquals(query_builder.to_query()['filter'], {
            'and': [
                {'exists': {'field': 'some_field'}},
                {'range': {'range_field': {'lte': 5}}}
            ]
        })

    def test_add_or_shorthand(self):
        query_builder = QueryBuilder()
        query_builder.or_filter('range', 'range_field', gte=5)
        self.assertEquals(query_builder.to_query()['filter'], {
            'or': [
                {'range': {'range_field': {'gte': 5}}}
            ]
        })

    def test_add_shorthand_negation(self):
        query_builder = QueryBuilder()
        query_builder.and_filter('~exists', 'some_field')
        self.assertEquals(query_builder.to_query()['filter'], {
            'and': [
                {'not': {'exists': {'field': 'some_field'}}},
            ]
        })

    def test_source_default(self):
        query_builder = QueryBuilder()
        self.assertEqual(query_builder.to_query()['_source'], True)

    def test_source_enable_disable(self):
        query_builder = QueryBuilder()

        query_builder.disable_source()
        self.assertEqual(query_builder.to_query()['_source'], False)

        query_builder.enable_source()
        self.assertEqual(query_builder.to_query()['_source'], True)

    def test_sort_default(self):
        query_builder = QueryBuilder()
        query_builder.sort('test')
        self.assertEqual(query_builder.to_query()['sort'], [
            {'test': 'asc'}
        ])

    def test_sort_desc(self):
        query_builder = QueryBuilder()
        query_builder.sort('test', 'desc')
        self.assertEqual(query_builder.to_query()['sort'], [
            {'test': 'desc'}
        ])

    def test_sort_asc(self):
        query_builder = QueryBuilder()
        query_builder.sort('test', 'asc')
        self.assertEqual(query_builder.to_query()['sort'], [
            {'test': 'asc'}
        ])

    def test_sort_reset(self):
        query_builder = QueryBuilder()
        query_builder.sort('test', 'asc')
        self.assertEqual(query_builder.to_query()['sort'], [
            {'test': 'asc'}
        ])
        query_builder.sort_reset()
        self.assertFalse('sort' in query_builder.to_query())

    @patch('es_fluent.filters._now')
    def test_age_filter(self, mock_now):
        mock_now.return_value = datetime.datetime(2015, 1, 1, 2, 0, 0)
        age = Age('some_field', lte=3600, gte=1800, gt=1860, lt=1920)
        self.assertEqual(age.to_query(), {
            'range': {
                'some_field': {
                    'gte': '2015-01-01T01:00:00',
                    'lte': '2015-01-01T01:30:00',
                    'lt': '2015-01-01T01:29:00',
                    'gt': '2015-01-01T01:28:00',
                }
            }
        })

    def test_missing_filter(self):
        missing = Missing('field_name')
        self.assertEqual(missing.to_query(), {
            "missing": {
                "field": "field_name"
            }
        })

    def test_size(self):
        query_builder = QueryBuilder()
        self.assertNotIn('size', query_builder.to_query())
        query_builder.size = 50
        self.assertIn('size', query_builder.to_query())
        self.assertEqual(query_builder.to_query()['size'], 50)


if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())



"""
Exercises the RegExp filter.
"""
import unittest
import re

from es_fluent.filters import RegExp


class TestEs_fluent_filters(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_invalid_regexp(self):
        with self.assertRaises(re.error):
            RegExp("field", r"[$")

    def test_regexp(self):
        re = RegExp(
            "field", r".*test_Exp$"
        )
        self.assertEquals(
            re.to_query(),
            {
                "regexp": {"field": ".*test_Exp$"}
            }
        )
